@@ -5,11 +5,10 @@
 
 package Future::Queue;
 
-use v5.10;
 use strict;
 use warnings;
 
-our $VERSION = '0.48_003';
+our $VERSION = '0.42';
 
 =head1 NAME
 
